# -*- coding: utf-8 -*-

"""\
Wrappers for multiple types of osu! related files.

Source: https://github.com/cmyui/cmyui_pkg

:copyright: (c) 2021 cmyui
:license: MIT, see LICENSE for details.
"""

from .beatmap import *
from .oppai_ng import *
from .replay import *
